import pytest
from pyplaces import foursquare_open_places as fsq, overture_maps as ov
from pyplaces._errors import PyArrowError,UnsupportedOperatorError


user_optional_valid = {"columns":["confidence","names"],"filters":("confidence",">",0.9),
                    "distance":1,"unit":"km"}
user_provided_valid = {"address":"204 Hemenway Street","place":"Jamaica Plain, MA","bbox":(-71.11088398563271, 42.32145496739517, -71.0987557723746, 42.33045192410487)}
user_optional_invalid = {"columns":["sfdsfd"],"filters_invalid_op":("confidence","(",0.9),"filters_invalid_field":("fdscds","<",0.9),
            "filters_invalid_value":("confidence", "<", "a"),"unit":"sfd","release":"2025-10-23"}
user_provided_invalid = {"address":"fdlksjfld","bbox":(-181,-91, 181, 91),"point":(-181, -91)}

def test_foursquare_places():
    
    places_schema = ['fsq_place_id', 'name', 'latitude', 'longitude', 'address', 'locality',
                    'region', 'postcode', 'admin_region', 'post_town', 'po_box', 'country',
                    'date_created', 'date_refreshed', 'date_closed', 'tel', 'website',
                    'email', 'facebook_id', 'instagram', 'twitter', 'fsq_category_ids',
                    'fsq_category_labels', 'placemaker_url', 'geom', 'bbox']
    places_schema.sort()
    categories_schema = ["category_id","category_level","category_name","category_label","level1_category_id","level1_category_name",
                            "level2_category_id","level2_category_name","level3_category_id","level3_category_name","level4_category_id",
                            "level4_category_name","level5_category_id","level5_category_name","level6_category_id","level6_category_name"]
    categories_schema.sort()
    
    #check that categories schema is correct
    
    with pytest.raises(ValueError):
        #assert invalid release
        fsq.foursquare_places_from_address(user_provided_valid["address"],release=user_optional_invalid["release"])
        #assert invalid unit
        fsq.foursquare_places_from_address(user_provided_valid["address"],unit=user_optional_invalid["unit"])
        
        
        #TODO not sure about these
        #assert invalid point
        fsq.foursquare_places_from_address(user_optional_invalid["point"])
        #assert invalid address
        fsq.foursquare_places_from_place(user_provided_invalid["address"])
        #assert invalid bbox
        fsq.foursquare_places_from_bbox(user_provided_invalid["bbox"])
    with pytest.raises(PyArrowError):
        #assert invalid column
        fsq.foursquare_places_from_address(user_provided_valid["address"],user_optional_invalid["columns"])
        #assert invalid field
        fsq.foursquare_places_from_address(user_provided_valid["address"],filters=user_optional_invalid["filters_invalid_field"])
        #assert invalid value
        fsq.foursquare_places_from_address(user_provided_valid["address"],filters=user_optional_invalid["filters_invalid_value"])
    with pytest.raises(UnsupportedOperatorError):
        #assert invalid operator
        fsq.foursquare_places_from_address(user_provided_valid["address"],filters=user_optional_invalid["filters_invalid_op"])

    #assert correct case, all valid inputs. check that schema is correct for each
    cols1=fsq.foursquare_places_from_address(user_provided_valid["address"]).columns.to_list()
    cols2 = fsq.foursquare_places_from_place(user_provided_valid["place"]).columns.to_list()
    cols3 = fsq.foursquare_places_from_bbox(user_provided_valid["bbox"]).columns.to_list()
    cols4 = fsq.get_categories().columns.to_list()
    cols1.sort()
    cols2.sort()
    cols3.sort()
    cols4.sort()
    assert cols1 == places_schema
    assert cols2 == places_schema
    assert cols3 == places_schema
    assert cols4 == categories_schema


def test_overture_maps():
    
    places_schema = ['id', 'geometry', 'bbox', 'version', 'sources', 'names', 'categories',
                            'confidence', 'websites', 'socials', 'emails', 'phones', 'brand','addresses']
    addresses_schema = ['id', 'geometry', 'bbox', 'country', 'postcode', 'street', 'number',
                        'unit', 'address_levels', 'postal_city', 'version', 'sources']
    buildings_schema = ['id', 'geometry', 'bbox', 'version', 'sources', 'level', 'subtype',
                            'class', 'height', 'names', 'has_parts', 'is_underground', 'num_floors',
                            'num_floors_underground', 'min_height', 'min_floor', 'facade_color',
                            'facade_material', 'roof_material', 'roof_shape', 'roof_direction',
                            'roof_orientation', 'roof_color', 'roof_height']
    buildings_part_schema = ['id', 'geometry', 'bbox', 'version', 'sources', 'level', 'height',
                                'names', 'is_underground', 'num_floors', 'num_floors_underground',
                                'min_height', 'min_floor', 'facade_color', 'facade_material',
                                'roof_material', 'roof_shape', 'roof_direction', 'roof_orientation',
                                'roof_color', 'roof_height', 'building_id']
    transportation_segment_schema = ['id', 'geometry', 'bbox', 'version', 'sources', 'subtype', 'class',
                                            'names', 'connectors', 'routes', 'subclass', 'subclass_rules',
                                            'access_restrictions', 'level_rules', 'destinations',
                                            'prohibited_transitions', 'road_surface', 'road_flags', 'speed_limits',
                                            'width_rules']
    transportation_connector_schema = ['id', 'geometry', 'bbox', 'version', 'sources']
    addresses_schema.sort()
    places_schema.sort()
    buildings_schema.sort()
    buildings_part_schema.sort()
    transportation_segment_schema.sort()
    transportation_connector_schema.sort()
    
    #assert correct schema for all of these
    cols1 = ov.overture_addresses_from_bbox(user_provided_valid["bbox"]).columns.to_list()
    # ov.overture_base_from_bbox(user_provided_valid["bbox"])
    cols2 = ov.overture_places_from_bbox(user_provided_valid["bbox"]).columns.to_list()
    #building and building part
    cols3 = ov.overture_buildings_from_bbox(user_provided_valid["bbox"]).columns.to_list()
    cols4 = ov.overture_buildings_from_bbox(user_provided_valid["bbox"],building_part=True).columns.to_list()
    #connector and segment
    cols5 = ov.overture_transportation_from_bbox(user_provided_valid["bbox"]).columns.to_list()
    cols6 = ov.overture_transportation_from_bbox(user_provided_valid["bbox"],connector=True).columns.to_list()
    
    cols1.sort()
    cols2.sort()
    cols3.sort()
    cols4.sort()
    cols5.sort()
    cols6.sort()
    assert cols1 == addresses_schema
    assert cols2 == places_schema
    assert cols3 == buildings_schema
    assert cols4 == buildings_part_schema
    assert cols5 == transportation_segment_schema
    assert cols6 == transportation_connector_schema


"""Core utility functions."""

from __future__ import annotations

import sys
import functools
import inspect
import re
from typing import Union, List, Tuple, Any, Literal
from typing_extensions import TypeAlias

from pyarrow.compute import field
from pyarrow.dataset import Expression

from ._errors import UnsupportedOperatorError, PyArrowError

def run_before_decorator(before_func):
    """Returns a decorator that runs the specified `before_func` before the wrapped function."""
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Extract 'release' argument dynamically
            sig = inspect.signature(func)
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()

            if "release" in bound_args.arguments:
                release_value = bound_args.arguments["release"]
                before_func(release_value)  # Call the module-specific before function

            return func(*args, **kwargs)

        return wrapper
    return decorator

def wrap_functions_with_release(module_name, before_func):
    """Dynamically wraps all functions with 'release' parameter in the given module."""
    module = sys.modules[module_name]
    
    for name in dir(module):
        attr = getattr(module, name)
        if callable(attr) and name != before_func.__name__:
            sig = inspect.signature(attr)
            if "release" in sig.parameters:  # Wrap only functions with "release"
                setattr(module, name, run_before_decorator(before_func)(attr))

FieldName: TypeAlias = str
OperatorStr: TypeAlias = Literal["==", "!=", "<", "<=", ">", ">=",
                                "is_nan", "is_null", "is_valid", "isin"]
FilterValue: TypeAlias = Union[str, int, float, List[Any], Tuple[Any, ...], None]
FilterTuple: TypeAlias = Tuple[FieldName, OperatorStr, FilterValue]
FilterGroup: TypeAlias = List[FilterTuple]

FilterStructure: TypeAlias = List[Union[FilterTuple, FilterGroup]] | FilterTuple
"""FilterStructure represents a list of filtering rules for a DataFrame-like object."""

def tuple_to_expression(filter_tuple: FilterTuple) -> Expression:
    """
    Convert a tuple of [field_name, operator, value] into a PyArrow expression.
    
    Parameters:
    -----------
    filter_tuple : tuple
        A tuple of (field_name, operator, value) where:
        - field_name (str): The name of the field to filter on
        - operator (str): The operator as a string ('>', '<', '==', etc.)
        - value: The value to compare against (str, int, or list)
    
    Returns:
    --------
    pyarrow.compute.Expression
        The resulting PyArrow expression
    """
    if len(filter_tuple) != 3:
        raise ValueError("Filter tuple must have exactly 3 elements: (field_name, operator, value)")
    
    field_name, op_str, value = filter_tuple
    pyaro_field = field(field_name)
    
    # Map operator strings to PyArrow operations    
    if op_str == "==" or op_str == "=":
        return pyaro_field == value
    elif op_str == "!=":
        return pyaro_field != value
    elif op_str == "<":
        return pyaro_field < value
    elif op_str == "<=":
        return pyaro_field <= value
    elif op_str == ">":
        return pyaro_field > value
    elif op_str == ">=":
        return pyaro_field >= value
    elif op_str == "is_nan":
        return pyaro_field.is_nan()
    elif op_str == "is_null":
        return pyaro_field.is_null()
    elif op_str == "is_valid":
        return pyaro_field.is_valid()
    elif op_str == "isin":
        if not isinstance(value, (list, tuple)):
            value = [value]
        return pyaro_field.isin(value)
    else:
        raise UnsupportedOperatorError(f"Unsupported operator: {op_str}")

def build_filter_expression(filter_structure: FilterStructure) -> Expression:
    """
    Build a PyArrow expression from a nested filter structure.
    
    Parameters:
    -----------
    filter_structure : list
        A list of filter groups, where each filter group is:
        - A single filter tuple (field_name, operator, value)
        - OR a list of filter tuples that should be combined with OR logic
    
    Returns:
    --------
    pyarrow.compute.Expression
        The resulting combined PyArrow expression
    """
    if not filter_structure:
        return None
    
    if not isinstance(filter_structure, list):
        filter_structure = [filter_structure]
    
    # Process each filter or filter group
    expressions = []
    
    for filter_item in filter_structure:
        if isinstance(filter_item, tuple):
            # Single filter tuple
            expressions.append(tuple_to_expression(filter_item))
        elif isinstance(filter_item, list):
            # List of OR'd filter tuples
            if not filter_item:
                continue  # Skip empty lists
                
            # Convert each tuple to expression
            or_expressions = [tuple_to_expression(ft) for ft in filter_item]
            
            # Combine with OR logic
            or_expr = or_expressions[0]
            for expr in or_expressions[1:]:
                or_expr = or_expr | expr  # Use '|' for logical OR
                
            expressions.append(or_expr)
    
    # Combine all expressions with AND logic
    combined_expr = expressions[0]
    for expr in expressions[1:]:
        combined_expr = combined_expr & expr  # Use '&' for logical AND
    
    return combined_expr

def catch_column_filter_error(exc_type: BaseException,error: Exception) -> None:
    """
    Throw user-friendly PyArrow errors.
    
    Parameters:
    -----------
    exc_type : BaseException
        Exception data
    error : Exception
        Exception thrown
    
    Returns:
    --------
    None
    """
    # Capture the full traceback
    error_message = str(error)
    # print(exc_type.__name__)
    if exc_type.__name__ == "UnsupportedOperatorError":
        raise error
    elif exc_type.__name__ == "ArrowInvalid":
        match = re.search(r"FieldRef\.Name\(([^)]+)\)", error_message)
        name = match.group(1)
        raise PyArrowError(f"Invalid column name:\"{name}\"") from error
    elif exc_type.__name__ =="ArrowNotImplementedError":
        match = re.search(r"\(([^)]+)\)", error_message)
        first_value,last_value = match.group(1).split(",")
        raise ValueError(f"Incorrect type used for value in filter: \"{last_value.strip()}\" should be \"{first_value.strip()}\"") from error


"""Unit conversion utilities."""

from __future__ import annotations

def convert_to_meters(value: float, unit: str) -> float:
    """
    Convert a value from the specified unit to meters.
    
    Parameters:
    -----------
    value : float
        The value to convert
    unit : str
        The source unit ('m', 'km', 'in', 'ft', 'yd', 'mi')
        
    Returns:
    --------
    float
        Value in meters
    
    Raises:
    -------
    ValueError
        If the unit is not recognized
    """
    conversion_factors = {
        "m": 1,          # meters
        "km": 1000,      # kilometers to meters
        "in": 0.0254,    # inches to meters
        "ft": 0.3048,    # feet to meters
        "yd": 0.9144,    # yards to meters
        "mi": 1609.34,   # miles to meters
    }
    
    if unit not in conversion_factors:
        raise ValueError(f"Invalid unit: {unit}. Valid units: m, km, in, ft, yd, mi")
    
    return value * conversion_factors[unit]


"""Geographic and geometry utility functions."""

from __future__ import annotations

from pyproj import Geod
from shapely import Polygon
from ._conversion_utils import convert_to_meters
from osmnx.geocoder import geocode, geocode_to_gdf
from osmnx import settings

# Set OSMNX settings
settings.http_referer = "pyplaces Python Package" 
settings.http_user_agent = "pyplaces"

def point_buffer(lon: float, lat: float, radius_m: float) -> Polygon:
    """
    Create a circular buffer around a point using the WGS84 ellipsoid.
    
    Parameters:
    -----------
    lon : float
        Longitude of the center point
    lat : float
        Latitude of the center point
    radius_m : float
        Radius of the buffer in meters
        
    Returns:
    --------
    shapely.Polygon
        Polygon representing the buffer
    """
    
    geod = Geod(ellps='WGS84')
    num_vtxs = 64
    
    # Generate equally spaced angles in degrees from 360 to 0
    angles = []
    for i in range(num_vtxs):
        angle = 360 - (i * 360 / num_vtxs)
        angles.append(angle)
    
    # Generate the coordinates for each vertex of the buffer
    coords = []
    for angle in angles:
        # Calculate a single point - geod.fwd returns lon, lat, az for a single point
        fwd_lon, fwd_lat, _ = geod.fwd(lon, lat, angle, radius_m, radians=False)
        coords.append((fwd_lon, fwd_lat))
    
    # Create a polygon from the coordinates
    return Polygon(coords)

def geocode_point_to_bbox(address: str, distance: float, unit: str):
    """
    Convert an address or coordinates to a bounding box using a buffer.
    
    Parameters:
    -----------
    address : str or tuple
        Address string or (longitude, latitude) tuple
    distance : float
        Buffer distance
    unit : str
        Unit of distance (m, km, ft, etc.)
        
    Returns:
    --------
    tuple
        Bounding box as (minx, miny, maxx, maxy)
    """
    if isinstance(address, str):
        point = geocode(address)
    else:
        point = address
    distance = convert_to_meters(distance, unit)
    bbox = point_buffer(point[1], point[0], distance).bounds
    return bbox

def geocode_place_to_bbox(address: str):
    """
    Convert a place name to its geometry and bounding box.
    
    Parameters:
    -----------
    address : str
        Place name or address
        
    Returns:
    --------
    tuple
        (geometry, bbox) where bbox is (minx, miny, maxx, maxy)
    """
    gdf = geocode_to_gdf(query=address, which_result=1, by_osmid=False)
    row = gdf.iloc[0]
    geometry = row["geometry"]
    bbox = (row["bbox_west"], row["bbox_south"], row["bbox_east"], row["bbox_north"])
    return geometry, bbox


# from .overture_maps import overture_addresses_from_address,overture_addresses_from_bbox,overture_addresses_from_place,overture_base_from_address, overture_base_from_bbox,overture_base_from_place,overture_buildings_from_address,overture_buildings_from_bbox,overture_buildings_from_place,overture_places_from_address,overture_places_from_bbox,overture_places_from_place,overture_transportation_from_address,overture_transportation_from_bbox,overture_transportation_from_place
# from .foursquare_open_places import foursquare_places_from_address, foursquare_places_from_bbox, foursquare_places_from_place, get_categories

__all__ = ['overture_maps']
__all__ = ['foursquare_open_places']


"""Functions to fetch  geoparquet data from Overture Maps on AWS"""
from geopandas import GeoDataFrame
from importlib import resources
from ._utils import FilterStructure, wrap_functions_with_release
from ._io_utils import from_address, from_bbox, from_place

#TODO latest release reads from text file
OVERTURE_MAIN_PATH = 's3://overturemaps-us-west-2/release/{release}/'
OVERTURE_BUCKET = 'overturemaps-us-west-2'
OVERTURE_REGION = 'us-west-2'
OVERTURE_LATEST_RELEASE = "2025-01-22.0"

OVERTURE_PLACES_PREFIX = "theme=places/type=place/"
OVERTURE_BUILDINGS_PREFIX = "theme=buildings/type=building/"
OVERTURE_BUILDINGS_PART_PREFIX = "theme=buildings/type=building_part/"
OVERTURE_ADDRESSES_PREFIX = "theme=addresses/type=address/"
OVERTURE_TRANSPORTATION_SEGMENT_PREFIX = "theme=transportation/type=segment/"
OVERTURE_TRANSPORTATION_CONNECTOR_PREFIX = "theme=transportation/type=connector/"
OVERTURE_BASE_PREFIX = "theme=base/type={type}"

def overture_places_from_address(address: str,columns: list[str]| None = None,filters: FilterStructure | None = None,distance: float = 500 ,unit: str = "m" ,release: str = OVERTURE_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieve places data from Overture based on an address.
    
    Parameters
    ----------
    address : str
        The address to search from
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filters to apply to the data, by default None
    distance : float, optional
        Search radius from the address, by default 500
    unit : str, optional
        Unit for the distance ('m' for meters, etc.), by default "m"
    release : str, optional
        Overture data release version, by default OVERTURE_LATEST_RELEASE
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing places data around the specified address
    """
    return from_address(address,OVERTURE_PLACES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters,distance,unit)

def overture_places_from_place(address: str,columns: list[str]| None=None,filters: FilterStructure| None=None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame:
    """
    Retrieve places data from Overture based on a named place.
    
    Parameters
    ----------
    address : str
        The place name to search within
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filters to apply to the data, by default None
    release : str, optional
        Overture data release version, by default OVERTURE_LATEST_RELEASE
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing places data within the specified place
    """
    return from_place(address,OVERTURE_PLACES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_places_from_bbox(bbox: tuple[float,float,float,float],columns: list[str]| None=None,filters: FilterStructure| None=None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame:
    """
    Retrieve places data from Overture within a bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        The bounding box coordinates (min_x, min_y, max_x, max_y)
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filters to apply to the data, by default None
    release : str, optional
        Overture data release version, by default OVERTURE_LATEST_RELEASE
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing places data within the specified bounding box
    """
    return from_bbox(bbox,OVERTURE_PLACES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_buildings_from_address(address: str,columns: list[str]| None = None,filters: FilterStructure| None = None,distance: float = 500 ,unit: str = "m" ,release: str = OVERTURE_LATEST_RELEASE,building_part: bool=False) -> GeoDataFrame:
    """
    Retrieve buildings data from Overture based on an address.
    
    Parameters
    ----------
    address : str
        The address to search from
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filters to apply to the data, by default None
    distance : float, optional
        Search radius from the address, by default 500
    unit : str, optional
        Unit for the distance ('m' for meters, etc.), by default "m"
    release : str, optional
        Overture data release version, by default OVERTURE_LATEST_RELEASE
    building_part : bool, optional
        Whether to retrieve building parts instead of whole buildings, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing buildings data around the specified address
    """
    if building_part:
        prefix = OVERTURE_BUILDINGS_PART_PREFIX
    else:
        prefix = OVERTURE_BUILDINGS_PREFIX 
    return from_address(address,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters,distance,unit)

def overture_buildings_from_place(address: str,columns: list[str]| None=None,filters: FilterStructure| None=None,release: str=OVERTURE_LATEST_RELEASE,building_part: bool=False)-> GeoDataFrame:
    """
    Retrieve buildings data from Overture based on a named place.
    
    Parameters
    ----------
    address : str
        The place name to search within
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filters to apply to the data, by default None
    release : str, optional
        Overture data release version, by default OVERTURE_LATEST_RELEASE
    building_part : bool, optional
        Whether to retrieve building parts instead of whole buildings, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing buildings data within the specified place
    """
    if building_part:
        prefix = OVERTURE_BUILDINGS_PART_PREFIX
    else:
        prefix = OVERTURE_BUILDINGS_PREFIX
    return from_place(address,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_buildings_from_bbox(bbox: tuple[float,float,float,float],columns: list[str]| None=None,filters: FilterStructure| None=None,release: str=OVERTURE_LATEST_RELEASE,building_part: bool=False)-> GeoDataFrame:
    """
    Retrieve buildings data from Overture within a bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        The bounding box coordinates (min_x, min_y, max_x, max_y)
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filters to apply to the data, by default None
    release : str, optional
        Overture data release version, by default OVERTURE_LATEST_RELEASE
    building_part : bool, optional
        Whether to retrieve building parts instead of whole buildings, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing buildings data within the specified bounding box
    """
    if building_part:
        prefix = OVERTURE_BUILDINGS_PART_PREFIX
    else:
        prefix = OVERTURE_BUILDINGS_PREFIX
    return from_bbox(bbox,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_transportation_from_address(address: str,columns: list[str]| None = None,filters: FilterStructure| None = None,distance: float = 500 ,unit: str = "m" ,release: str = OVERTURE_LATEST_RELEASE,connector:bool=False) -> GeoDataFrame:
    """
    Retrieve transportation data from Overture based on an address.
    
    Parameters
    ----------
    address : str
        The address to search from
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filters to apply to the data, by default None
    distance : float, optional
        Search radius from the address, by default 500
    unit : str, optional
        Unit for the distance ('m' for meters, etc.), by default "m"
    release : str, optional
        Overture data release version, by default OVERTURE_LATEST_RELEASE
    connector : bool, optional
        Whether to retrieve connectors instead of segments, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing transportation data around the specified address
    """
    if connector:
        prefix = OVERTURE_TRANSPORTATION_CONNECTOR_PREFIX
    else:
        prefix = OVERTURE_TRANSPORTATION_SEGMENT_PREFIX
    return from_address(address,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters,distance,unit)

def overture_transportation_from_place(address: str,columns: list[str]| None=None,filters: FilterStructure| None=None,release: str=OVERTURE_LATEST_RELEASE,connector:bool=False)-> GeoDataFrame:
    """
    Retrieve transportation data from Overture based on a named place.
    
    Parameters
    ----------
    address : str
        The place name to search within
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filters to apply to the data, by default None
    release : str, optional
        Overture data release version, by default OVERTURE_LATEST_RELEASE
    connector : bool, optional
        Whether to retrieve connectors instead of segments, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing transportation data within the specified place
    """
    if connector:
        prefix = OVERTURE_TRANSPORTATION_CONNECTOR_PREFIX
    else:
        prefix = OVERTURE_TRANSPORTATION_SEGMENT_PREFIX
    return from_place(address,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_transportation_from_bbox(bbox: tuple[float,float,float,float],columns: list[str]| None=None,filters: FilterStructure | None=None,release: str=OVERTURE_LATEST_RELEASE,connector:bool=False)-> GeoDataFrame:
    """
    Retrieve transportation data from Overture within a bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        The bounding box coordinates (min_x, min_y, max_x, max_y)
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filters to apply to the data, by default None
    release : str, optional
        Overture data release version, by default OVERTURE_LATEST_RELEASE
    connector : bool, optional
        Whether to retrieve connectors instead of segments, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing transportation data within the specified bounding box
    """
    if connector:
        prefix = OVERTURE_TRANSPORTATION_CONNECTOR_PREFIX
    else:
        prefix = OVERTURE_TRANSPORTATION_SEGMENT_PREFIX
    return from_bbox(bbox,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)
    
def overture_addresses_from_address(address: str,columns: list[str]| None = None,filters: FilterStructure| None = None,distance: float = 500 ,unit: str = "m" ,release: str = OVERTURE_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieve address data from Overture based on an address.
    
    Parameters
    ----------
    address : str
        The address to search from
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filters to apply to the data, by default None
    distance : float, optional
        Search radius from the address, by default 500
    unit : str, optional
        Unit for the distance ('m' for meters, etc.), by default "m"
    release : str, optional
        Overture data release version, by default OVERTURE_LATEST_RELEASE
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing address data around the specified address
    """
    return from_address(address,OVERTURE_ADDRESSES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters,distance,unit)

def overture_addresses_from_place(address: str,columns: list[str]| None=None,filters: FilterStructure| None=None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame:
    """
    Retrieve address data from Overture based on a named place.
    
    Parameters
    ----------
    address : str
        The place name to search within
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filters to apply to the data, by default None
    release : str, optional
        Overture data release version, by default OVERTURE_LATEST_RELEASE
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing address data within the specified place
    """
    return from_place(address,OVERTURE_ADDRESSES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_addresses_from_bbox(bbox: tuple[float,float,float,float],columns: list[str]| None=None,filters: FilterStructure| None=None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame:
    """
    Retrieve address data from Overture within a bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        The bounding box coordinates (min_x, min_y, max_x, max_y)
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filters to apply to the data, by default None
    release : str, optional
        Overture data release version, by default OVERTURE_LATEST_RELEASE
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing address data within the specified bounding box
    """
    return from_bbox(bbox,OVERTURE_ADDRESSES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_base_from_address(address: str,base_type: str,columns: list[str]| None = None,filters: FilterStructure| None = None,distance: float = 500 ,unit: str = "m" ,release: str = OVERTURE_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieve base data of a specific type from Overture based on an address.
    
    Parameters
    ----------
    address : str
        The address to search from
    base_type : str
        The type of base data to retrieve
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filters to apply to the data, by default None
    distance : float, optional
        Search radius from the address, by default 500
    unit : str, optional
        Unit for the distance ('m' for meters, etc.), by default "m"
    release : str, optional
        Overture data release version, by default OVERTURE_LATEST_RELEASE
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing base data of the specified type around the address
    """
    _check_base_type(base_type)
    complete_prefix = OVERTURE_BASE_PREFIX.format(base_type=base_type)
    return from_address(address,complete_prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters,distance,unit)
    

def overture_base_from_place(address: str,base_type: str,columns: list[str]| None=None,filters: FilterStructure=None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame:
    """
    Retrieve base data of a specific type from Overture based on a named place.
    
    Parameters
    ----------
    address : str
        The place name to search within
    base_type : str
        The type of base data to retrieve
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure, optional
        Filters to apply to the data, by default None
    release : str, optional
        Overture data release version, by default OVERTURE_LATEST_RELEASE
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing base data of the specified type within the place
    """
    _check_base_type(base_type)
    complete_prefix = OVERTURE_BASE_PREFIX.format(base_type=base_type)
    return from_place(address,complete_prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)
    

def overture_base_from_bbox(bbox: tuple[float,float,float,float],base_type: str,columns: list[str]| None=None,filters: FilterStructure| None=None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame  :
    """
    Retrieve base data of a specific type from Overture within a bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        The bounding box coordinates (min_x, min_y, max_x, max_y)
    base_type : str
        The type of base data to retrieve
    columns : list[str] | None, optional
        Specific columns to include in the result, by default None
    filters : FilterStructure | None, optional
        Filters to apply to the data, by default None
    release : str, optional
        Overture data release version, by default OVERTURE_LATEST_RELEASE
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing base data of the specified type within the bounding box
    """
    _check_base_type(base_type)
    complete_prefix = OVERTURE_BASE_PREFIX.format(base_type=base_type)
    return from_bbox(bbox,complete_prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)



#TODO no automation
def _check_base_type(base_type):
    """
    Validate if the provided base type is valid.
    
    Parameters
    ----------
    base_type : str
        The base type to validate
        
    Raises
    ------
    ValueError
        If the base type is not valid
    """   
    with open("releases/overture/base_types.txt", "r",encoding="utf-8-sig") as f:
        folders = [line.replace("type=", "").strip(" \n/") for line in f]
    if base_type not in folders:
        raise ValueError(f"Invalid base type:{base_type}")
    
def _check_release(release):
    """
    Validate if the provided Overture release is valid.
    
    Parameters
    ----------
    release : str
        The release version to validate
        
    Raises
    ------
    ValueError
        If the release version is not valid
    """
    with resources.files("pyplaces").joinpath("releases/overture/releases.txt").open( "r",encoding="utf-8-sig") as f:
        folders = [line.strip(" \n/") for line in f]
    if release not in folders:
        raise ValueError(f"Invalid release:{release}")
    
wrap_functions_with_release(__name__, _check_release)

__all__ = ["overture_addresses_from_address","overture_addresses_from_bbox","overture_addresses_from_place","overture_base_from_address", 
            "overture_base_from_bbox","overture_base_from_place","overture_buildings_from_address","overture_buildings_from_bbox",
            "overture_buildings_from_place","overture_places_from_address","overture_places_from_bbox","overture_places_from_place",
            "overture_transportation_from_address","overture_transportation_from_bbox","overture_transportation_from_place"]



"""Functions to fetch  geoparquet data from Foursquare Open Places on AWS"""
from importlib import resources
from geopandas import GeoDataFrame
from pandas import DataFrame
from ._utils import FilterStructure, wrap_functions_with_release
from ._io_utils import from_address, from_bbox, from_place, read_parquet_arrow

#TODO latest release reads from text file
FSQ_MAIN_PATH = 's3://fsq-os-places-us-east-1/release/dt={release}/'
FSQ_BUCKET = 'fsq-os-places-us-east-1'
FSQ_REGION = 'us-east-1'
FSQ_PLACES_PREFIX = "places/parquet/"
FSQ_CATEGORIES_PREFIX = "categories/parquet/"
FSQ_LATEST_RELEASE = "2025-02-06"

FSQ_FUSED_MAIN_PATH = 's3://us-west-2.opendata.source.coop/fused/fsq-os-places/{release}/'
FSQ_FUSED_BUCKET = 's-west-2.opendata.source.coop'
FSQ_FUSED_REGION = 'us-west-2'
FSQ_FUSED_LATEST_RELEASE = "2025-01-10"
FSQ_FUSED_PLACES_PREFIX = "places/"

def foursquare_places_from_address(address: str, 
                                    columns: list[str] | None = None, 
                                    filters: FilterStructure | None = None, 
                                    distance: float = 500, 
                                    unit: str = "m", 
                                    release: str = FSQ_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieves Foursquare places data around a specified address.
    
    Parameters
    ----------
    address : str
        The address to search for nearby places.
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset. If None, all columns are retrieved.
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results.
    distance : float, default 500
        Radius of the search area around the address.
    unit : str, default "m"
        Unit of measurement for the distance (e.g., "m" for meters, "km" for kilometers).
    release : str, default FSQ_LATEST_RELEASE
        Foursquare data release version to use.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing Foursquare places data within the specified distance of the address.
    """
    return from_address(address, FSQ_PLACES_PREFIX, FSQ_MAIN_PATH, FSQ_REGION, release, columns, filters, distance, unit)


def foursquare_places_from_place(address: str, 
                                    columns: list[str] | None = None, 
                                    filters: FilterStructure = None, 
                                    release: str = FSQ_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieves Foursquare places data for a specific place identified by its address.
    
    Parameters
    ----------
    address : str
        The address or identifier of the place to retrieve data for.
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset. If None, all columns are retrieved.
    filters : FilterStructure, optional
        Filter criteria to apply to the results.
    release : str, default FSQ_LATEST_RELEASE
        Foursquare data release version to use.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing Foursquare places data for the specified place.
    """
    return from_place(address, FSQ_PLACES_PREFIX, FSQ_MAIN_PATH, FSQ_REGION, release, columns, filters)


def foursquare_places_from_bbox(bbox: tuple[float, float, float, float], 
                                columns: list[str] | None = None, 
                                filters: FilterStructure | None = None, 
                                release: str = FSQ_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieves Foursquare places data within a specified bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        Bounding box coordinates in the format (min_x, min_y, max_x, max_y).
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset. If None, all columns are retrieved.
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results.
    release : str, default FSQ_LATEST_RELEASE
        Foursquare data release version to use.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing Foursquare places data within the specified bounding box.
    """
    return from_bbox(bbox, FSQ_PLACES_PREFIX, FSQ_MAIN_PATH, FSQ_REGION, release, columns, filters)


def get_categories(columns: list[str] | None = None, 
                    filters: FilterStructure | None = None, 
                    release: str = FSQ_LATEST_RELEASE) -> DataFrame:
    """
    Retrieves Foursquare place categories data.
    
    Parameters
    ----------
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset. If None, all columns are retrieved.
    filters : FilterStructure | None, optional
        Filter criteria to apply to the results.
    release : str, default FSQ_LATEST_RELEASE
        Foursquare data release version to use.
        
    Returns
    -------
    DataFrame
        A DataFrame containing Foursquare place categories data.
        
    Notes
    -----
    This function constructs the data path using Foursquare-specific constants and
    retrieves the categories data using read_parquet_arrow.
    """
    path = FSQ_MAIN_PATH.format(release=release) + FSQ_CATEGORIES_PREFIX
    return read_parquet_arrow(path, FSQ_REGION, columns, filters)


def _check_release(release: str):
    """
    Validates if the specified Foursquare data release version exists.
    
    Parameters
    ----------
    release : str
        The release version to validate.
        
    Raises
    ------
    ValueError
        If the specified release version does not exist in the available releases.
        
    Notes
    -----
    This private helper function reads available release versions from a text file
    and checks if the specified release is among them.
    """
    with resources.files("pyplaces").joinpath("releases/foursquare/releases.txt").open("r", encoding="utf-8-sig") as f:
        folders = [line.replace("dt=", "").strip(" \n/") for line in f]
    if release not in folders:
        raise ValueError(f"Invalid release: {release}")
    
wrap_functions_with_release(__name__, _check_release)
    
        # from pyarrow.parquet import ParquetFile
    # from pyarrow.fs import S3FileSystem
    # from rapidfuzz import process

    # # S3 Configuration
    # S3_BUCKET = "your-bucket-name"
    # S3_KEY = "path/to/categories.parquet"


    # def find_matching_foursquare_categories(user_input, s3_path, threshold=60, batch_size=1000):
    #     """Streams the Parquet file in chunks and performs fuzzy matching"""
    #     s3_filesystem = S3FileSystem()
    #     # Open the Parquet file
    #     parquet_file = ParquetFile(s3_path, filesystem=s3_filesystem)
        
    #     # Read in chunks
    #     for batch in parquet_file.iter_batches(batch_size, columns=["category_name", "category_id"]):
    #         df = batch.to_pandas()  # Convert batch to Pandas DataFrame
    #         categories_dict = dict(zip(df["category_name"], df["category_id"]))  # Convert to dict
            
    #         # Perform fuzzy matching
    #         matches = process.extract(user_input, categories_dict.keys(), limit=5, score_cutoff=threshold)
    #         if matches:
    #             return {name: categories_dict[name] for name, score, _ in matches}
        
    #     return {}  # Return empty dict if no matches

    # # Example usage
    # user_query = input("Enter category: ").strip()
    # s3_path = f"{S3_BUCKET}/{S3_KEY}"
    # matching_categories = find_matching_categories(user_query, s3_path, s3_filesystem)

    # if matching_categories:
    #     print("Matched Categories and IDs:", matching_categories)
    # else:
    #     print("No relevant categories found.")


__all__ = ["foursquare_places_from_address", "foursquare_places_from_bbox", "foursquare_places_from_place", "get_categories"]


"""I/O utility functions for reading data."""

from __future__ import annotations

import functools
import operator
from json import loads
import sys

from pyarrow import RecordBatchReader
from pyarrow.compute import field
from pyarrow.dataset import dataset
from pyarrow.fs import S3FileSystem
from geopandas import GeoDataFrame

from ._utils import build_filter_expression, catch_column_filter_error, FilterStructure
from ._geo_utils import geocode_place_to_bbox, geocode_point_to_bbox
from ._errors import S3ReadError

def decode_bytes(obj):
    """Recursively decode byte strings in nested structures."""
    if isinstance(obj, dict):
        return {decode_bytes(k): decode_bytes(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [decode_bytes(i) for i in obj]
    elif isinstance(obj, bytes):
        return obj.decode("utf-8")
    else:
        return obj

def read_geoparquet_arrow(path: str, region: str, bbox: tuple[float,float,float,float], 
                        columns: list[str] | None = None, 
                        filters: FilterStructure | None = None) -> GeoDataFrame:
    """
    Read geospatial data from a parquet file on S3 with filtering by bbox.
    
    Parameters:
    -----------
    path : str
        S3 path to the parquet file
    region : str
        AWS region
    bbox : tuple
        Bounding box as (minx, miny, maxx, maxy)
    columns : list, optional
        Columns to select
    filters : FilterStructure, optional
        Filter expression
        
    Returns:
    --------
    GeoDataFrame
        Filtered geodataframe
    """
    filter_expr = build_filter_expression(filters)
    
    xmin, ymin, xmax, ymax = bbox
    
    geo_filter_expr = (
        (field("bbox", "xmin") < xmax)
        & (field("bbox", "xmax") > xmin)
        & (field("bbox", "ymin") < ymax)
        & (field("bbox", "ymax") > ymin)
    )
    filter_ls = list(filter(lambda x: x is not None, [geo_filter_expr, filter_expr]))
    combined_filter = functools.reduce(operator.and_, filter_ls)
    
    clean_path = path.replace("s3://", "")
    try:
        ds = dataset(
            clean_path, filesystem=S3FileSystem(anonymous=True, region=region)
        )
    except Exception as e:
        raise S3ReadError(f"Read from bucket {clean_path} could not be complete.") from e
    
    try:
        if columns:
            batches = ds.to_batches(columns=columns, filter=combined_filter)
        else:
            batches = ds.to_batches(filter=combined_filter)
    except Exception as e:
        exc_info = sys.exc_info()[0]
        catch_column_filter_error(exc_info,e)
    
    non_empty_batches = (b for b in batches if b.num_rows > 0)
    
    schema = ds.schema
    metadata_str = decode_bytes(schema.metadata) 
    geo_dict = loads(metadata_str["geo"])
    geo_column = geo_dict["primary_column"]
    
    geometry_field_index = schema.get_field_index(geo_column)
    geometry_field = schema.field(geometry_field_index)
    geoarrow_geometry_field = geometry_field.with_metadata(
        {b"ARROW:extension:name": b"geoarrow.wkb"}
    )

    geoarrow_schema = schema.set(geometry_field_index, geoarrow_geometry_field)
    reader = RecordBatchReader.from_batches(geoarrow_schema, non_empty_batches)

    return GeoDataFrame.from_arrow(reader)

def read_parquet_arrow(path: str, region: str, 
                    columns: list[str] | None = None, 
                    filters: FilterStructure | None = None) -> GeoDataFrame:
    """
    Read tabular data from a parquet file on S3.
    
    Parameters:
    -----------
    path : str
        S3 path to the parquet file
    region : str
        AWS region
    columns : list, optional
        Columns to select
    filters : FilterStructure, optional
        Filter expression
        
    Returns:
    --------
    GeoDataFrame
        Filtered dataframe
    """
    filter_expr = build_filter_expression(filters)
    clean_path = path.replace("s3://", "")
    ds = dataset(
        clean_path, filesystem=S3FileSystem(anonymous=True, region=region)
    )
    try:
        if columns and filter_expr:
            batches = ds.to_batches(columns=columns, filter=filter_expr)
        elif filter_expr:
            batches = ds.to_batches(filter=filter_expr)
        else:
            batches = ds.to_batches(columns=columns)
    except Exception as e:
        exc_info = sys.exc_info()[0]
        catch_column_filter_error(exc_info,e)
    
    non_empty_batches = (b for b in batches if b.num_rows > 0)
    schema = ds.schema
    reader = RecordBatchReader.from_batches(schema, non_empty_batches)
    return reader.read_pandas()

def get_gdf_from_bbox(release:str, bbox:tuple[float,float,float,float], columns:list[str], filters: FilterStructure, prefix: str, path: str, region: str):
    """Helper function to get a geodataframe from a bounding box."""
    main_path = path.format(release=release) + prefix
    gdf = read_geoparquet_arrow(main_path, region, bbox, columns=columns, filters=filters)
    return gdf

def from_address(address: str | tuple[float,float], prefix: str, main_path: str, region: str,
            release: str, columns: list[str]| None = None, filters: FilterStructure| None = None,
            distance: float = 500, unit: str = "m") -> GeoDataFrame:
    """
    Wrapper to geocode an address and fetch the geoparquet data within the address's area.
    
    Parameters:
    -----------
    address : str or tuple
        Address string or (longitude, latitude) tuple
    prefix : str
        Path prefix for the data
    main_path : str
        Base path template with {release} placeholder
    region : str
        AWS region
    release : str
        Release version
    columns : list, optional
        Columns to select
    filters : FilterStructure, optional
        Filter expression
    distance : float, default 500
        Buffer distance
    unit : str, default 'm'
        Unit of distance
        
    Returns:
    --------
    GeoDataFrame
        Filtered geodataframe
    """
    bbox = geocode_point_to_bbox(address, distance, unit)
    gdf = from_bbox(bbox,prefix,main_path,region,release,columns,filters)
    return gdf
    
def from_place(address: str, prefix: str, main_path: str, region: str, release: str,
            columns: list[str]| None=None, filters: FilterStructure| None=None) -> GeoDataFrame:
    """
    Wrapper to geocode a place and fetch the geoparquet data within the place.
    
    Parameters:
    -----------
    address : str
        Place name or address
    prefix : str
        Path prefix for the data
    main_path : str
        Base path template with {release} placeholder
    region : str
        AWS region
    release : str
        Release version
    columns : list, optional
        Columns to select
    filters : FilterStructure, optional
        Filter expression
        
    Returns:
    --------
    GeoDataFrame
        Filtered geodataframe
    """
    geometry, bbox = geocode_place_to_bbox(address)
    gdf = from_bbox(bbox,prefix,main_path,region,release,columns,filters)
    filtered_gdf = gdf[gdf.within(geometry)]
    return filtered_gdf

def from_bbox(bbox: tuple[float,float,float,float], prefix: str, main_path: str, region: str, 
            release: str, columns: list[str]| None=None, filters: FilterStructure| None=None) -> GeoDataFrame:
    """
    Wrapper to fetch the geoparquet data within the bounding box.
    
    Parameters:
    -----------
    bbox : tuple
        Bounding box as (minx, miny, maxx, maxy)
    prefix : str
        Path prefix for the data
    main_path : str
        Base path template with {release} placeholder
    region : str
        AWS region
    release : str
        Release version
    columns : list, optional
        Columns to select
    filters : FilterStructure, optional
        Filter expression
        
    Returns:
    --------
    GeoDataFrame
        Filtered geodataframe
    """
    gdf = get_gdf_from_bbox(release, bbox, columns, filters, prefix, main_path, region)
    return gdf


from pyplaces.foursquare_open_places import foursquare_places_from_address


# foursquare_places_from_address("204 Hemenway Street",filters=[("incorrect_column",">",0.9)])

# places.overture_addresses_from_address("204 Hemenway Street",filters=[("confidence",">","incorrect_type")])

# df = pyplaces.overture_maps.overture_places_from_address("204 Hemenway Street")

df = foursquare_places_from_address("204 Hemenway Street")

print(df.columns)
print(df.head(10))


class UnsupportedOperatorError(NotImplementedError):
    """Exception for catching unsupported PyArrow operators"""

class S3ReadError(FileNotFoundError):
    """Exception for empty results when querying an S3 bucket."""

class PyArrowError(ValueError):
    """Exception for catching PyArrow errors when invalid parameters are passed."""